/*
** main.c
**
** Nokia 3310 Example Program
** Target: ATMEGA :: AVR-GCC
**
** Written by Tony Myatt - 2007
** Quantum Torque - www.quantumtorque.com
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include "lcd.h"
#include <util/delay.h>
#include <string.h>

#define BOARDSIZE 3
#define XPLAYER 'x'
#define OPLAYER 'o'
#define INF 100
#define WIN 10

#define BLANK '-'

#define TEST

static char board[BOARDSIZE][BOARDSIZE];
char msj[30];

static inline int evaluate(char player){
	int i;
	for(i=0;i<BOARDSIZE;++i){
		if(board[i][0] == player && board[i][0] == board[i][1] && board[i][1] == board[i][2])
			return WIN;
		if(board[0][i] == player && board[0][i] == board[1][i] && board[1][i] == board[2][i])
			return WIN;
	}
	if(board[0][0] == player && board[0][0] == board[1][1] && board[1][1] == board[2][2])
		return WIN;
	if(board[2][0] == player && board[2][0] == board[1][1] && board[1][1] == board[0][2])
		return WIN;	
	return 0;
}

static int8_t minimax_x(char player,char opp,uint8_t depth){
	uint8_t i,j;
	int8_t max= - INF,score;
	score = evaluate(opp);
	if(score || depth == 0)
		return -score;
	for(i=0;i<BOARDSIZE;++i){
		for(j=0;j<BOARDSIZE;++j){
			if(board[i][j] == BLANK){
				board[i][j] = player;
				score = -minimax_x(opp,player,depth - 1);
				board[i][j] = BLANK;
				if(score>max){
					max = score;
				}
			}
		}
	}
	return max;
}

static inline uint8_t minimax_0(char player,char opp,uint8_t depth){
	uint8_t i,j,poz;
	int8_t max= - INF,score;
	for(i=0;i<BOARDSIZE;++i){
		for(j=0;j<BOARDSIZE;++j){
			if(board[i][j] == BLANK){
				board[i][j] = player;
				score = -minimax_x(opp,player,depth-1);
				board[i][j] = BLANK;
				if(score>max){
					max = score;
					poz = i*BOARDSIZE+j;
				}
			}
		}
	}
	return poz;
}

static inline void printBoard(int x,int y,char*aux){
	uint8_t i,j;
	LCD_clear();
	for(i=0;i<BOARDSIZE;++i){
		LCD_gotoXY(0,i);
		for(j=0;j<BOARDSIZE;++j){
			if(i==y && j==x){
				LCD_chr(board[i][j],SELECTED);
			}
			else{
				LCD_chr(board[i][j],NOT_SELECTED);
			}
		}
	}
	
	LCD_gotoXY(0,5);
	LCD_str(aux,0);
}

static int getKeyPres(int*x,int*y){
	if(~BUTTONPIN & BUTTONRIGHT){
		++*x;
		if(*x>=BOARDSIZE)
			*x=0;
	}
	else if(~BUTTONPIN & BUTTONLEFT){
		--*x;
		if(*x<0)
			*x+=BOARDSIZE;
	}
	else if(~BUTTONPIN & BUTTONUP){
		++*y;
		if(*y>=BOARDSIZE)
			*y=0;
	}
	else if(~BUTTONPIN & BUTTONDOWN){
		--*y;
		if(*y<0)
			*y+=BOARDSIZE;
	}
	else if(~BUTTONPIN & BUTTONSELECT){
		if(board[*y][*x] == BLANK){
			board[*y][*x] = 'x';
			#ifdef TEST
			_delay_ms(5);
			#else
			_delay_ms(500);
			#endif
			return 1;
		}
		else{
			strcpy(msj,"Va rugam sa selectati o casuta goala!");
		}
	}
	#ifdef TEST
	_delay_ms(5);
	#else
	_delay_ms(500);
	#endif
	return 0;
}


static void play(){
	uint8_t i,j;
	int x=0,y=0;
	int8_t scor = 0;
	int mutare;
	uint8_t depth = BOARDSIZE * BOARDSIZE;
	for(i=0;i<BOARDSIZE;++i){
		for(j=0;j<BOARDSIZE;++j){
			board[i][j] = BLANK;
		}
	}
	printBoard(x,y,msj);
	do{
		strcpy(msj,"Pozitionati x");
		do{
			i = getKeyPres(&x,&y);
			printBoard(x,y,msj);
		}while(i==0);
		printBoard(x,y,msj);
		--depth;
		scor = evaluate(XPLAYER);
		if(scor == 0 && depth){
			mutare = minimax_0(OPLAYER,XPLAYER,depth);
			y = mutare/BOARDSIZE;
			x = mutare%BOARDSIZE;
			sprintf(msj,"o pe (%d %d)",y+1,x+1);
			board[y][x] = OPLAYER;
			--depth;
			scor = -evaluate(OPLAYER);
			printBoard(x,y,msj);
			#ifdef TEST
			_delay_ms(5);
			#else
			_delay_ms(1000);
			#endif
		}
	}while((scor==0) && depth);
	if(scor<0){
		sprintf(msj,"Am castigat!");
	}
	else if(scor<0){
		sprintf(msj,"Ai castigat!");
	}
	else{
		sprintf(msj,"Remiza!");
	}
	printBoard(x,y,msj);
	_delay_ms(3000);
}

/* main -- Program starts here */
int main(void){
	// Setup LCD
	LCD_init();
	LCD_contrast(0x40);
	
	BUTTONPORT = 0xff;
	BUTTONDDR  = 0;
	
	while(1)
		play();
}
